from yta_editor_nodes_cpu.processor.abstract import _NodeProcessorCPU, _NodeProcessorCoreCPU
from yta_editor_utils.texture import TextureUtils
from typing import Union

import numpy as np


class ColorContrastNodeProcessorCPU(_NodeProcessorCPU):
    """
    The node to modify the color contrast of an input
    by using the CPU.

    This class can be instantiated many different
    times with different parameters, but will always
    call the same `Singleton` node processor instance
    to process the `input`.
    """

    mandatory_inputs = ['base_input']
    optional_inputs = []
    
    def __init__(
        self,
        cpu_render_context: 'CPURenderContext'
    ):
        super().__init__(
            node_processor = _ColorContrastNodeProcessorCoreCPU(),
            cpu_render_context = cpu_render_context
        )

    def process(
        self,
        inputs: dict[str, np.ndarray],
        evaluation_context: 'EvaluationContext',
        factor: float = 1.5,
        **kwargs
    ) -> np.ndarray:
        """
        Process the `input` provided by using the node
        processor instance associated to this node and
        return the output.
        """
        return self.node_processor.process(
            input = inputs['base_input'],
            output_size = self.cpu_render_context.output_size,
            factor = factor
        )
    
class _ColorContrastNodeProcessorCoreCPU(_NodeProcessorCoreCPU):
    """
    *For internal use only*
    
    Node processor to modify the color contrast of the
    input.

    Depending on the factor:
    - `factor>1.0` = More contrast
    - `factor==1.0` = Same contrast
    - `factor<1.0` = Less contrast

    TODO: Improve this, its just temporary
    """

    def process(
        self,
        input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        factor: float,
        **kwargs
    ):
        # Ensure float32 [0, 1]
        source = TextureUtils.numpy_to_float32(input)

        # We do this to be the same as in GPU but it
        # should be calculated... and also for the GPU
        mean = np.array([0.5, 0.5, 0.5, 0.5], dtype = np.float32)

        # mean = input.mean(
        #     axis = (0, 1),
        #     keepdims = True
        # )
        """
        factor > 1 -> More contrast
        factor < 1 -> Less contrast
        """
        out = (source - mean) * factor + mean
        out = np.clip(out, 0, 1)

        return out


from yta_editor_nodes_cpu.processor.abstract import _NodeProcessorCPU, _NodeProcessorCoreCPU
from abc import abstractmethod
from typing import Union

import numpy as np


class _TransitionProcessorCPU(_NodeProcessorCPU):
    """
    *For internal use only*

    A transition between the frames of 2 videos.

    This transition is made with CPU (numpy).
    """

    mandatory_inputs = ['first_input', 'second_input']
    optional_inputs = []

    def __init__(
        self,
        node_processor: '_TransitionProcessorCoreCPU',
        cpu_render_context: 'CPURenderContext'
    ):
        # TODO: Validate '_TransitionProcessorCoreCPU' or child
        super().__init__(
            node_processor = node_processor,
            cpu_render_context = cpu_render_context
        )

    def process(
        self,
        inputs: dict[str, np.ndarray],
        evaluation_context: 'EvaluationContext',
        **kwargs
    ) -> np.ndarray:
        """
        Process the `first_input` and `second_input` and
        generate the transition frame according to the
        `progress` of the transition provided.
        """
        return self.node_processor.process(
            first_input = inputs['first_input'],
            second_input = inputs['second_input'],
            output_size = self.cpu_render_context.output_size,
            progress = float(evaluation_context.progress),
            **kwargs
        )
    

class _TransitionProcessorCoreCPU(_NodeProcessorCoreCPU):
    """
    *For internal use only*

    *Singleton class*

    Class to represent a transition in between 2
    different videos. This class will be called
    internally by the specific nodes to process
    the inputs.
    """

    @abstractmethod
    def process(
        self,
        first_input: np.ndarray,
        second_input: np.ndarray,
        progress: float,
        **kwargs
    ) -> np.ndarray:
        """
        Process the `first_input` and `second_input` and
        generate the transition frame according to the
        `progress` of the transition provided.
        """
        # TODO: Specific attributes can be received as
        # **kwargs to modify the specific process
        pass
    


from yta_editor_nodes_cpu.processor.audio.abstract import _AudioNodeProcessorCoreCPU, _AudioNodeProcessorCPU

import numpy as np


class ChorusAudioNodeProcessorCPU(_AudioNodeProcessorCPU):
    """
    The audio node processor to apply a chorus
    effect.
    """

    def __init__(
        self,
        sample_rate: int,
        depth: int = 0,
        frequency: float = 0.25
    ):
        # TODO: Describe these attributes, please
        self.sample_rate: int = sample_rate
        self.depth: int = depth
        self.frequency: float = frequency

        super().__init__(
            audio_node_processor = _ChorusAudioNodeProcessorCoreCPU()
        )

    def process(
        self,
        input: np.ndarray,
        t: float,
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided audio 'input' that
        is played on the given 't' time moment.
        """
        return self.audio_node_processor.process(
            input = input,
            t = t,
            sample_rate = self.sample_rate,
            depth = self.depth,
            frequency = self.frequency
        )
    
class _ChorusAudioNodeProcessorCoreCPU(_AudioNodeProcessorCoreCPU):
    """
    *For internal use only*

    *Singleton class*

    The audio node processor to apply a chorus
    effect.
    """

    def process(
        self,
        input: np.ndarray,
        t: float,
        sample_rate: int,
        depth: int,
        frequency: float,
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided audio 'input' that
        is played on the given 't' time moment.
        """
        n_samples = input.shape[0]
        t = np.arange(n_samples) / sample_rate

        # Sinusoidal LFO that controls the delay
        delay = (depth / 1000.0) * sample_rate * (0.5 * (1 + np.sin(2 * np.pi * frequency * t)))
        delay = delay.astype(np.int32)

        output = np.zeros_like(input, dtype=np.float32)

        for i in range(n_samples):
            d = delay[i]

            output[i]= (
                0.7 * input[i] + 0.7 * input[i - d]
                if (i - d) >= 0 else
                input[i]
            )

        return output


from yta_editor_nodes_common import _Node

import numpy as np


class _NodeCPU(_Node):
    """
    *For internal use only*
    
    *Abstract class*

    An abstract class to handle the mandatory and
    optional parameters that can be used in the node.
    """

    mandatory_inputs = []
    optional_inputs = []

    def _validate_inputs(
        self,
        inputs: dict[str, np.ndarray]
    ) -> None:
        return super()._validate_inputs(
            inputs = inputs,
            type = np.ndarray
        )


from yta_editor_nodes_cpu.processor.video.transitions.abstract import _TransitionProcessorCoreCPU, _TransitionProcessorCPU
from typing import Union

import numpy as np


class SlideTransitionProcessorCPU(_TransitionProcessorCPU):
    """
    A transition in which the frames goes from one
    side to the other, disappearing the first one
    and appearing the second one.
    """

    mandatory_inputs = ['first_input', 'second_input']
    optional_inputs = []

    def __init__(
        self,
        cpu_render_context: 'CPURenderContext'
    ):
        super().__init__(
            node_processor = _SlideTransitionProcessorCoreCPU(),
            cpu_render_context = cpu_render_context
        )

class _SlideTransitionProcessorCoreCPU(_TransitionProcessorCoreCPU):
    """
    *For internal use only*

    *Singleton class*

    A transition in which the frames goes from one
    side to the other, disappearing the first one
    and appearing the second one.
    """

    def process(
        self,
        # TODO: What about the type (?)
        first_input: np.ndarray,
        second_input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        progress: float,
        **kwargs
    ) -> np.ndarray:
        """
        Process the `first_input` and `second_input` and
        generate the transition frame according to the
        `progress` of the transition provided.
        """
        # TODO: Should we keep this functionality (?)
        # first_input = (
        #     texture_to_frame(first_input, do_include_alpha = True)
        #     if PythonValidator.is_instance_of(first_input, 'moderngl.Texture') else
        #     first_input
        # )

        # second_input = (
        #     texture_to_frame(second_input, do_include_alpha = True)
        #     if PythonValidator.is_instance_of(second_input, 'moderngl.Texture') else
        #     second_input
        # )

        # TODO: What do we do in this case (?)
        assert first_input.shape == second_input.shape, 'Los frames deben tener la misma forma'

        h, w, c = first_input.shape
        frame_out = np.zeros_like(first_input)

        # By default we are only handling 'left' direction
        direction = 'left'

        offset_x = 0
        # offset_y = 0
        if direction == 'left':
            offset_x = int(w * progress)
            if offset_x < w:
                frame_out[:, :w - offset_x] = first_input[:, offset_x:]
            if offset_x > 0:
                frame_out[:, w - offset_x:] = second_input[:, :offset_x]

        # elif direction == 'right':
        #     offset_x = int(w * progress)
        #     if offset_x < w:
        #         frame_out[:, offset_x:] = first_input[:, :w - offset_x]
        #     if offset_x > 0:
        #         frame_out[:, :offset_x] = second_input[:, w - offset_x:]

        # elif direction == 'up':
        #     offset_y = int(h * progress)
        #     if offset_y < h:
        #         frame_out[:h - offset_y, :] = first_input[offset_y:, :]
        #     if offset_y > 0:
        #         frame_out[h - offset_y:, :] = second_input[:offset_y, :]

        # elif direction == 'down':
        #     offset_y = int(h * progress)
        #     if offset_y < h:
        #         frame_out[offset_y:, :] = first_input[:h - offset_y, :]
        #     if offset_y > 0:
        #         frame_out[:offset_y, :] = second_input[h - offset_y:, :]

        return frame_out


from yta_editor_nodes_cpu.compositor.abstract import _NodeCompositorCoreCPU, _NodeCompositorCPU
from typing import Union

import numpy as np


class DisplacementWithRotationNodeCompositorCPU(_NodeCompositorCPU):
    """
    TODO: Explain this

    TODO: This must be implemented
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []

    def __init__(
        self,
        cpu_render_context: 'CPURenderContext'
    ):
        super().__init__(
            node_compositor = _DisplacementWithRotationNodeCompositorCoreCPU(),
            cpu_render_context = cpu_render_context
        )

"""
TODO: Using Pillow is very slow, so we will not
implement this by now. But here is some code that
should work once it's modified
"""
class _DisplacementWithRotationNodeCompositorCoreCPU(_NodeCompositorCoreCPU):
    """
    *For internal use only*

    TODO: This must be implemented
    """

    def process(
        self,
        base_input: np.ndarray,
        overlay_input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        **kwargs
    ) -> np.ndarray:
        raise NotImplementedError('Not implemented yet')
    
# TODO: This one is a bit special because you need
# more than 1 texture, and it is also the one we
# use to move the frames, so move it from here
# class DisplacementWithRotationNodeCompositorCPU(_NodeCompositorCPU):
#     """
#     The frame, but moving and rotating over other frame.
#     """

#     def process(
#         self,
#         # TODO: What about the type (?)
#         base_input: np.ndarray,
#         overlay_input: np.ndarray,
#         position: tuple[int, int],
#         size: tuple[int, int],
#         rotation: int,
#         **kwargs
#     ):
#         from yta_numpy.utils import numpy_to_pillow

#         # TODO: Maybe 'cv2' is faster (?)
#         base_im = numpy_to_pillow(
#             frame = base_input,
#             do_read_as_rgba = True
#         )
#         ovr_im = numpy_to_pillow(
#             frame = overlay_input,
#             do_read_as_rgba = True
#         )

#         base_w, base_h = base_im.size

#         pos_x, pos_y = position
#         tgt_w = int(round(size[0]))
#         tgt_h = int(round(size[1]))

#         # Avoid 0 size
#         tgt_w = max(1, tgt_w)
#         tgt_h = max(1, tgt_h)

#         from PIL import Image

#         # Resize overlay to expected size
#         overlay_resized = ovr_im.resize((tgt_w, tgt_h), resample = Image.BICUBIC)

#         # --- Rotar alrededor del centro (expand=False mantiene bbox tamaño tgt_w x tgt_h) ---
#         # Pillow rota CCW; expand=False mantiene el canvas, recortando los bordes que sobresalgan.
#         overlay_rot = overlay_resized.rotate(rotation_degrees, resample=resample, expand=False)

#         # --- Crear capa vacía del tamaño de la base y pegar overlay en su posición centrada ---
#         overlay_layer = Image.new('RGBA', (base_w, base_h), (0, 0, 0, 0))

#         top_left_x = int(round(pos_x - tgt_w / 2.0))
#         top_left_y = int(round(pos_y - tgt_h / 2.0))

#         # Pegar usando la propia alpha de overlay_rot como máscara; paste maneja recortes fuera de bordes correctamente.
#         overlay_layer.paste(overlay_rot, (top_left_x, top_left_y), overlay_rot)

#         # --- Componer: base sobre la que overlay_layer se coloca ---
#         # Usamos alpha_composite pero requiere imágenes del mismo tamaño (ya lo son)
#         result = Image.alpha_composite(base_im, overlay_layer)

#         return np.array(result, dtype=np.uint8)


from yta_editor_nodes_cpu.processor.abstract import _NodeProcessorCPU, _NodeProcessorCoreCPU
from yta_editor_utils.texture import TextureUtils
from typing import Union

import numpy as np


class BlackAndWhiteNodeProcessorCPU(_NodeProcessorCPU):
    """
    The node to modify the input to become black and
    white only.

    This class can be instantiated many different
    times with different parameters, but will always
    call the same `Singleton` node processor instance
    to process the `input`.
    """

    mandatory_inputs = ['base_input']
    optional_inputs = []
    
    def __init__(
        self,
        cpu_render_context: 'CPURenderContext'
    ):
        super().__init__(
            node_processor = _BlackAndWhiteNodeProcessorCoreCPU(),
            cpu_render_context = cpu_render_context
        )

    def process(
        self,
        inputs: dict[str, np.ndarray],
        evaluation_context: 'EvaluationContext',
        # TODO: Should we set the 'factor' here to be able
        # to modify it if it is different than None (?)
        **kwargs
    ) -> np.ndarray:
        """
        Process the `input` provided by using the node
        processor instance associated to this node and
        return the output.
        """
        return self.node_processor.process(
            input = inputs['base_input'],
            output_size = self.cpu_render_context.output_size,
            **kwargs
        )
    
class _BlackAndWhiteNodeProcessorCoreCPU(_NodeProcessorCoreCPU):
    """
    *For internal use only*
    
    Node processor that modifies the input to become a
    black and white output.

    TODO: Improve this, its just temporary
    """

    def process(
        self,
        input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        **kwargs
    ) -> np.ndarray:
        """
        The `input` must be `float32` and the output will
        be also `float32`.
        """
        # Ensure float32 [0, 1]
        source = TextureUtils.numpy_to_float32(input)

        # Standard luminance conversion Rec.709
        gray = 0.2126 * source[..., 0] + 0.7152 * source[..., 1] + 0.0722 * source[..., 2]
        gray = np.clip(gray, 0, 1)

        return np.dstack(
            [gray, gray, gray, input[..., 3]]
            if input.shape[-1] == 4 else
            [gray, gray, gray]
        )
    



"""
We don't name it 'blend' instead of 'process' because
the GPU classes have the 'process' name due to the
inheritance from OpenGL classes, and we want to avoid
misunderstandings.
"""
from yta_editor_nodes_cpu.blender.add import AddBlenderCPU
from yta_editor_nodes_cpu.blender.alpha import AlphaBlenderCPU
from yta_editor_nodes_cpu.blender.mix import MixBlenderCPU


__all__ = [
    'AddBlenderCPU',
    'AlphaBlenderCPU',
    'MixBlenderCPU'
]


from yta_programming.singleton import SingletonABCMeta
from abc import ABC, abstractmethod

import numpy as np


class _AudioNodeProcessorCPU(ABC):
    """
    *For internal use only*

    *Abstract class*

    Node class, which is to implement specific
    parameters when instantiated and also a specific
    audio node processor that will process the
    `input` provided for a `t` time moment.
    """

    def __init__(
        self,
        audio_node_processor: '_AudioNodeProcessorCoreCPU'
    ):
        # TODO: Validate that is '_AudioNodeProcessorCoreCPU' or child
        self.audio_node_processor: '_AudioNodeProcessorCoreCPU' = audio_node_processor
        """
        The singleton instance of the processor that will
        process the input and return the output.
        """

    def process(
        self,
        input: np.ndarray,
        t: float,
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided audio `input` that
        is played on the given `t` time moment.
        """
        return self.audio_node_processor.process(
            input = input,
            t = t,
            **kwargs
        )

class _AudioNodeProcessorCoreCPU(metaclass = SingletonABCMeta):
    """
    *For internal use only*

    *Singleton class*

    The processor core that will be called by the
    specific audio node.
    """

    @abstractmethod
    def process(
        self,
        input: np.ndarray,
        t: float,
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided audio `input` that
        is played on the given `t` time moment.
        """
        pass


"""
When working with audio frames, we don't need
to use the GPU because audios are 1D and the
information can be processed perfectly with
a library like numpy.

If we need a very intense calculation for an
audio frame (FFT, convolution, etc.) we can
use CuPy or some DPS specific libraries, but
90% is perfectly done with numpy.

If you want to modify huge amounts of audio
(some seconds at the same time), you can use
CuPy, that has the same API as numpy but
working in GPU. Doing this below most of the
changes would work:
- `import numpy as np` → `import cupy as np`
"""
from yta_editor_nodes_cpu.processor.audio.volume import VolumeAudioNodeProcessorCPU


__all__ = [
    'VolumeAudioNodeProcessorCPU'
]


class CPURenderContext:
    """
    Class to wrap the information we need to
    be able to execute the nodes in CPU.

    This class must be created by the general
    node (located in `yta-editor-nodes` with the
    data from the general `RenderContext`).
    """

    def __init__(
        self,
        output_size: tuple[int, int]
    ):
        self.output_size: tuple[int, int] = output_size
        """
        The expected size for the output.
        """


from yta_editor_nodes_cpu.processor.audio.abstract import _AudioNodeProcessorCoreCPU, _AudioNodeProcessorCPU

import numpy as np


class VolumeAudioNodeProcessorCPU(_AudioNodeProcessorCPU):
    """
    The audio node processor to adjust the volume
    of the audio provided.
    """

    def __init__(
        self,
        # TODO: Maybe we need a `RateFunction` or similar
        factor_fn: callable
    ):
        self.factor_fn: callable = factor_fn
        """
        The function to calculate the `factor` to apply
        for a specific `t` time moment.
        """
        super().__init__(
            audio_node_processor = _VolumeAudioNodeProcessorCoreCPU()
        )

    def process(
        self,
        input: np.ndarray,
        t: float,
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided audio 'input' that
        is played on the given 't' time moment.
        """
        return self.audio_node_processor.process(
            input = input,
            t = t,
            factor_fn = self.factor_fn
        )

class _VolumeAudioNodeProcessorCoreCPU(_AudioNodeProcessorCoreCPU):
    """
    *For internal use only*

    *Singleton class*

    The audio node processor to adjust the volume
    of the audio provided.
    """

    def process(
        self,
        input: np.ndarray,
        t: float,
        # TODO: Maybe we need a `RateFunction` or similar
        factor_fn: callable,
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided audio 'input' that is
        played on the given 't' time moment and adjust
        the volume by using the `factor_fn` provided.
        """
        # TODO: This is not ok, change it
        factor = factor_fn(t, 0)

        samples = input
        samples *= factor

        # Determine dtype according to format
        # samples = (
        #     samples.astype(np.int16)
        #     # 'fltp', 's16', 's16p'
        #     if 's16' in input.format.name else
        #     samples.astype(np.float32)
        # )

        return samples



from yta_editor_nodes_cpu.processor.abstract import _NodeProcessorCPU, _NodeProcessorCoreCPU
from abc import abstractmethod
from typing import Union

import numpy as np


class _VideoNodeProcessorCPU(_NodeProcessorCPU):
    """
    *Abstract class*

    Node class, which is to implement specific
    parameters when instantiated and also a specific
    node processor that will process the `input` for
    the `t` time moment provided, because this is to
    affect a frame of a specitic moment of a video.
    """

    mandatory_inputs = ['base_input']
    optional_inputs = []

    def __init__(
        self,
        node_processor: '_VideoNodeProcessorCoreCPU',
        cpu_render_context: 'CPURenderContext'
    ):
        # TODO: Validate that is '_VideoNodeProcessorCoreCPU'
        super().__init__(
            node_processor = node_processor,
            cpu_render_context = cpu_render_context
        )

    def process(
        self,
        inputs: dict[str, np.ndarray],
        evaluation_context: 'EvaluationContext',
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided `input` and transform it by
        using the code that is defined here according to
        the `t` time moment provided.
        """
        return self.node_processor.process(
            input = inputs['base_input'],
            output_size = self.cpu_render_context.output_size,
            t = float(evaluation_context.t)
        )

class _VideoNodeProcessorCoreCPU(_NodeProcessorCoreCPU):
    """
    *Singleton class*

    *For internal use only*

    Class to represent a node processor that uses CPU
    to transform the input but it is meant to video
    processing, so it implements a 'time' parameter to
    manipulate the frames according to that time
    moment. This class will be called internally by the
    specific nodes to process the inputs.
    """

    # TODO: Just code and the same attributes that the
    # GPU version also has
    @abstractmethod
    def process(
        self,
        input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        t: float,
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided `input` and transform it by
        using the code that is defined here according to
        the `t` time moment provided.
        """
        # TODO: Specific attributes can be received as
        # **kwargs to modify the specific process
        pass


from yta_editor_nodes_cpu.complex.abstract import _NodeComplexCoreCPU, _NodeComplexCPU

import numpy as np


class DisplayOverAtNodeComplexCPU(_NodeComplexCPU):
    """
    The overlay input is placed in the scene with the
    given position, rotation and size, and then put as
    an overlay of the also given base input.

    Information:
    - The scene size is `(1920, 1080)`, so provide the
    `position` parameter according to it, where it is
    representing the center of the texture.
    - The `rotation` is in degrees, where `rotation=90`
    means rotating 90 degrees to the right.
    - The `size` parameter must be provided according to
    the previously mentioned scene size `(1920, 1080)`.
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []

    def __init__(
        self,
        cpu_render_context: 'CPURenderContext'
    ):
        super().__init__(
            node_complex = _DisplayOverAtNodeComplexCoreCPU(),
            cpu_render_context = cpu_render_context
        )

    def process(
        self,
        inputs: dict[str, np.ndarray],
        evaluation_context: 'EvaluationContext',
        position: tuple[int, int] = (1920 / 2, 1080 / 2),
        size: tuple[int, int] = (1920 / 2, 1080 / 2),
        rotation: int = 0,
        **kwargs
    ) -> np.ndarray:
        return self.node_complex.process(
            base_input = inputs['base_input'],
            overlay_input = inputs['overlay_input'],
            output_size = self.cpu_render_context.output_size,
            position = position,
            rotation = rotation,
            size = size
        )

class _DisplayOverAtNodeComplexCoreCPU(_NodeComplexCoreCPU):
    """
    *For internal use only*

    The overlay input is placed in the scene with the
    given position, rotation and size, and then put as
    an overlay of the also given base input.

    Information:
    - The scene size is `(1920, 1080)`, so provide the
    `position` parameter according to it, where it is
    representing the center of the texture.
    - The `rotation` is in degrees, where `rotation=90`
    means rotating 90 degrees to the right.
    - The `size` (of the element to display) parameter
    must be provided according to the previously
    mentioned scene size `(1920, 1080)`.

    TODO: This has no inheritance, is special and we need
    to be able to identify it as a valid one.

    TODO: This must be implemented
    """

    def process(
        self,
        base_input: np.ndarray,
        overlay_input: np.ndarray,
        output_size: tuple[int, int] = (1920, 1080),
        position: tuple[int, int] = (1920 / 2, 1080 / 2),
        size: tuple[int, int] = (1920 / 2, 1080 / 2),
        rotation: int = 0,
        **kwargs
    ) -> np.ndarray:
        raise NotImplementedError('Not implemented yet')


from yta_editor_nodes_cpu.processor.abstract import _NodeProcessorCPU, _NodeProcessorCoreCPU
from yta_editor_utils.texture import TextureUtils
from typing import Union

import numpy as np


class BrightnessNodeProcessorCPU(_NodeProcessorCPU):
    """
    The node to modify the brightness of an input
    by using the CPU.

    This class can be instantiated many different
    times with different parameters, but will always
    call the same `Singleton` node processor instance
    to process the `input`.
    """

    mandatory_inputs = ['base_input']
    optional_inputs = []
    
    def __init__(
        self,
        cpu_render_context: 'CPURenderContext'
    ):
        super().__init__(
            node_processor = _BrightnessNodeProcessorCoreCPU(),
            cpu_render_context = cpu_render_context
        )

    def process(
        self,
        inputs: dict[str, np.ndarray],
        evaluation_context: 'EvaluationContext',
        factor: float = 1.0,
        **kwargs
    ) -> np.ndarray:
        """
        Process the `input` provided by using the node
        processor instance associated to this node and
        return the output.
        """
        return self.node_processor.process(
            input = inputs['base_input'],
            output_size = self.cpu_render_context.output_size,
            factor = factor
        )
    
class _BrightnessNodeProcessorCoreCPU(_NodeProcessorCoreCPU):
    """
    *For internal use only*

    *Singleton class*

    Node processor to modify the brightness of the
    input by multiplying it by a factor.

    Depending on the factor:
    - `factor>1.0` = Lighter
    - `factor==1.0` = Same brightness
    - `factor<1.0` = Darker
    """

    def process(
        self,
        input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        factor: float,
        **kwargs
    ):
        """
        The `input` must be `float32` and the output will
        be also `float32`.
        """
        # Ensure float32 [0, 1]
        source = TextureUtils.numpy_to_float32(input)

        # Copy to avoid mutating upstream buffers
        output = source.copy()

        # Apply brightness (RGB only)
        output[..., :3] *= factor

        # Clamp
        output = np.clip(output, 0.0, 1.0)

        return output


"""
TODO: This module is duplicated as it is repeated
in the 'yta-editor-nodes-gpu' library. It should
be in a 'yta-editor-nodes-common' library.
"""
from yta_validation.parameter import ParameterValidator
from yta_validation.number import NumberValidator
from yta_validation import PythonValidator
from typing import Union

import numpy as np


def _validate_inputs_and_mix_weights(
    inputs: list[np.ndarray],
    mix_weights: Union[list[float], float] = 1.0,
) -> None:
    """
    *For internal use only*
    
    Validate the provided 'inputs' and 'mix_weights' raising
    an exception if something wrong was found.
    """
    ParameterValidator.validate_mandatory_list_of_these_instances('inputs', inputs, [np.ndarray, 'moderngl.Texture'])

    if len(inputs) < 2:
        raise Exception('The amount of "inputs" provided is less than 2.')
    
    if not all(input.size == inputs[0].size for input in inputs):
        # TODO: Should we raise this exception (?)
        raise Exception('All the inputs do not have the same size.')

    # Validate 'mix_weights' as a single or list of floats
    # in the [0.0, 1.0] range
    if PythonValidator.is_list_of_float(mix_weights):
        if len(mix_weights) != (len(inputs) - 1):
            raise Exception('The number of "mix_weights" values provided is not the same as "inputs".')
        
        for mix_weight in mix_weights:
            ParameterValidator.validate_mandatory_number_between('mix_weight', mix_weight, 0.0, 1.0)
    elif NumberValidator.is_number_between(mix_weights, 0.0, 1.0):
        mix_weights = [mix_weights] * len(inputs)
    else:
        raise Exception('The "mix_weights" parameter provided is not valid.')
    
    
def _clip_ndarray(
    input: np.ndarray,
    min: int = 0,
    max: int = 255
) -> np.ndarray:
    """
    *For internal use only*

    Clip the provided `input` numpy array with the also
    given `min` and `max` limits.
    """
    return np.clip(
        a = input,
        a_min = min,
        a_max = max
    )

def _get_stacked(
    inputs: list[np.ndarray]
) -> np.ndarray:
    """
    *For internal use only*

    Get the `inputs` but stacked.
    """
    return np.stack(
        arrays = inputs,
        axis = 0
    )

def _ensure_uint8(
    array: np.ndarray
) -> np.ndarray:
    """
    Ensure that the 'array' numpy array provided is in
    the uint8 [0, 255] range.

    This method will clip the array once it's been
    transformed (if necessary) and forced to be uint8.
    """
    if array.dtype == np.uint8:
        return array

    array = array.astype(np.float32)

    array = (
        array * 255.0
        # Detect normalized [0.0, 1.0] or in [0, 255]
        if array.max() <= 1.0 else
        array
    )

    return _clip_ndarray(array).astype(np.uint8)




from yta_programming.decorators.requires_dependency import requires_dependency
from yta_validation.parameter import ParameterValidator

import numpy as np


class InputHandler:
    """
    Class with static methods to simplify the way
    we handle the inputs to make them have the same
    size, able to be used in the nodes.
    """

    # TODO: Untested
    @staticmethod
    def pad(
        input: np.ndarray,
        size: tuple[int, int]
    ) -> np.ndarray:
        """
        Pad the `input` provided to fit the given `size`
        by adding transparent pixels around it and
        leaving the original input centered.
        """
        if input.ndim != 3:
            raise Exception(f'Expected shape (H, w, C), got {input.shape}')
        
        height, width, number_of_channels = input.shape

        if (width, height) > size:
            raise Exception('The input is bigger than the "size" requested.')
        elif (width, height) == size:
            return input
        
        # --- Normalizar a RGBA ---
        if number_of_channels == 3:
            alpha = np.full(
                shape = (height, width, 1),
                fill_value = 255,
                dtype = input.dtype
            )
            a_rgba = np.concatenate(
                arrays = [input, alpha],
                axis = 2
            )
        elif number_of_channels == 4:
            a_rgba = input
        else:
            raise Exception(f'Unexpected number of channels: {str(number_of_channels)}')

        # Transparent canvas
        canvas = np.zeros(
            shape = (size[1], size[0], 4),
            dtype = input.dtype
        )

        # --- Calcular offsets para centrar ---
        y0 = (size[1] - height) // 2
        x0 = (size[0] - width) // 2

        # --- Insertar imagen ---
        canvas[y0:y0 + height, x0:x0 + width] = a_rgba

        return canvas
    
    @requires_dependency('PIL', 'yta_numpy', 'pillow')
    def scale_to_biggest(
        inputs: list[np.ndarray]
    ) -> list[np.ndarray]:
        """
        Prepare all the `inputs` provided to have the same
        size (the size of the biggest one) by scalling them
        using pillow (`PIL`).
        """
        from yta_numpy.utils import scale_numpy_pillow

        ParameterValidator.validate_mandatory_list_of_these_instances('inputs', inputs, np.ndarray)

        # Look for the biggest size
        max_width = max(input.shape[1] for input in inputs)
        max_height = max(input.shape[0] for input in inputs)
        max_size: tuple[int, int] = (max_width, max_height)

        return [
            (
                # Scale if needed
                scale_numpy_pillow(input, max_size)
                if (input.shape[1], input.shape[0]) < max_size else
                input
            )
            for input in inputs
        ]



from yta_editor_nodes_cpu.blender.abstract import _BlenderCoreCPU, _BlenderCPU
from yta_editor_utils.texture import TextureUtils
from typing import Union

import numpy as np


class AddBlenderCPU(_BlenderCPU):
    """
    Blend the different inputs by applying the
    normal and most famous alpha blending and the
    most typical used in video editing.

    This blender will increase the brightness by
    combining the colors of the base and the overlay
    inputs, using the overlay as much as the 
    `stregth` parameter is indicating.
    """

    def __init__(
        self,
        cpu_render_context: 'CPURenderContext'
    ):
        super().__init__(
            blender = _AddBlenderCoreCPU(),
            cpu_render_context = cpu_render_context
        )

class _AddBlenderCoreCPU(_BlenderCoreCPU):
    """
    *Singleton class*

    *For internal use only*

    Blend the different inputs by applying the
    normal and most famous alpha blending and the
    most typical used in video editing.

    This blender will increase the brightness by
    combining the colors of the base and the overlay
    inputs, using the overlay as much as the 
    `stregth` parameter is indicating.
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []
    
    def process(
        self,
        inputs: dict[str, np.ndarray],
        output_size: Union[tuple[int, int], None],
        mix_weight: float = 1.0,
        dtype: Union[np.dtype, None] = None,
        strength: float = 1.0,
        **kwargs
    ) -> np.ndarray:
        # TODO: Maybe obtain the 'output_size' from
        # the 'base_input' if not defined...
        return super().process(
            inputs = inputs,
            output_size = output_size,
            mix_weight = mix_weight,
            dtype = dtype,
            strength = strength
        )
    
    def _blend(
        self,
        base_input: np.ndarray,
        overlay_input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        strength: float = 1.0,
        **kwargs
    ) -> np.ndarray:
        """
        *For internal use only*

        The internal process to blend and mix the provided
        `base_input` and `overlay_input`.

        This method should not force uint8 nor [0, 255]
        range by itself as it would be done in the 'blend'
        main method.
        """
        # We force to have float precission for the calculations
        base_input = TextureUtils.numpy_to_float32(base_input)
        overlay_input = TextureUtils.numpy_to_float32(overlay_input)

        # Add blending (base + overlay * strength)
        result = base_input + overlay_input * strength

        return TextureUtils.numpy_to_uint8(np.clip(result, 0.0, 1.0))


from yta_editor_nodes_cpu.abstract import _NodeCPU
from yta_programming.singleton import SingletonABCMeta
from abc import abstractmethod
from typing import Union

import numpy as np


class _NodeComplexCPU(_NodeCPU):
    """
    *For internal use only*

    *Abstract class*

    A node that is made by applying different nodes. It
    is different than the other nodes because it needs
    to import different nodes to process the input(s),
    by using CPU.
    """

    mandatory_inputs = ['base_input']
    optional_inputs = []

    def __init__(
        self,
        node_complex: '_NodeComplexCoreCPU',
        cpu_render_context: 'CPURenderContext'
    ):
        # TODO: Validate that is 'NodeComplexCoreCPU' or child
        self.node_complex: '_NodeComplexCoreCPU' = node_complex
        """
        The singleton instance of the complex node that
        will process the input and return the output.
        """
        self.cpu_render_context: 'CPURenderContext' = cpu_render_context
        """
        The render context for this CPU node processor.
        """

    def process(
        self,
        inputs: dict[str, np.ndarray],
        evaluation_context: 'EvaluationContext',
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided 'input' and transform it by
        using the code that is defined here.
        """
        return self.node_complex.process(
            input = inputs['base_input'],
            output_size = self.cpu_render_context.output_size,
            **kwargs
        )
    
class _NodeComplexCoreCPU(metaclass = SingletonABCMeta):
    """
    *For internal use only*

    *Singleton class*

    This class is to be able to use different nodes with
    the same input.

    This class will be called internally by the specific
    nodes to make the composition.
    """

    @abstractmethod
    def process(
        self,
        input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided 'input' and transform it by
        using the code that is defined here.
        """
        pass


from yta_editor_nodes_cpu.blender.abstract import _BlenderCoreCPU, _BlenderCPU
from typing import Union

import numpy as np


class MixBlenderCPU(_BlenderCPU):
    """
    Blend the second input with the first one
    applying the float factor (between 0.0 and
    1.0) that is passed as parameter.
    """

    def __init__(
        self,
        cpu_render_context: 'CPURenderContext'
    ):
        super().__init__(
            blender = _MixBlenderCoreCPU(),
            cpu_render_context = cpu_render_context
        )

class _MixBlenderCoreCPU(_BlenderCoreCPU):
    """
    *Singleton class*

    *For internal use only*

    Blend the second input with the first one
    applying the float factor (between 0.0 and
    1.0) that is passed as parameter.
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []

    def _blend(
        self,
        base_input: np.ndarray,
        overlay_input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        mix_weight: float = 1.0,
        **kwargs
    ) -> np.ndarray:
        """
        *For internal use only*

        The internal process to blend and mix the provided
        `base_input` and `overlay_input`.

        This method should not force uint8 nor [0, 255]
        range by itself as it would be done in the 'blend'
        main method.
        """
        # We do nothing because the mix is done in the
        # general 'process' method that calls this one
        # TODO: Maybe obtain the 'output_size' from
        # the 'overlay_input' if not defined...
        return overlay_input


"""
The nodes that modify inputs by using static parameters
and not 't' time moments, with CPU.
"""
from yta_editor_nodes_cpu.processor.black_and_white import BlackAndWhiteNodeProcessorCPU
from yta_editor_nodes_cpu.processor.brightness import BrightnessNodeProcessorCPU
from yta_editor_nodes_cpu.processor.color_contrast import ColorContrastNodeProcessorCPU
from yta_editor_nodes_cpu.processor.selection_mask import SelectionMaskNodeProcessorCPU
from yta_editor_nodes_cpu.processor.transparency import TransparencyNodeProcessorCPU


__all__ = [
    'BlackAndWhiteNodeProcessorCPU',
    'BrightnessNodeProcessorCPU',
    'ColorContrastNodeProcessorCPU',
    'SelectionMaskNodeProcessorCPU',
    'TransparencyNodeProcessorCPU'
]
    



# TODO: Should we keep this functionality (?)
# from yta_video_opengl.utils import texture_to_frame
from yta_editor_nodes_cpu.abstract import _NodeCPU
from yta_editor_nodes_cpu.blender.utils import _validate_inputs_and_mix_weights, _ensure_uint8
from yta_editor_nodes_cpu.utils import InputHandler
from yta_programming.singleton import SingletonABCMeta
from yta_validation.parameter import ParameterValidator
from abc import abstractmethod
from typing import Union

import numpy as np


class _BlenderCPU(_NodeCPU):
    """
    *For internal use only*

    Class to represent a blender that uses CPU to
    blend the inputs.

    This class must be implemented by any blender
    that uses CPU to blend inputs.
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []

    def __init__(
        self,
        blender: '_BlenderCoreCPU',
        cpu_render_context: 'CPURenderContext'
    ):
        self.blender: '_BlenderCoreCPU' = blender
        """
        The blender singleton instance that will be used
        internally to blend the inputs.
        """
        self.cpu_render_context: 'CPURenderContext' = cpu_render_context
        """
        The render context for this CPU node processor.
        """

    # TODO: Should this be in the '_BlenderCPU' instead (?)
    def process(
        self,
        inputs: dict[str, np.ndarray],
        evaluation_context: 'EvaluationContext',
        mix_weight: float = 1.0,
        dtype: Union[np.dtype, None] = None,
        **kwargs
    ) -> np.ndarray:
        """
        Blend the `base_input` provided with the also given
        `overlay_input`, using the `mix_weight` to obtain
        a single output as the result.

        The inputs should be received in uint8 format in
        the [0, 255] range, then processed with the specific
        processing method, and then turned back again to the
        uint8 format and [0, 255] range.
        """
        ParameterValidator.validate_dict_of('inputs', inputs, np.ndarray)

        # Validate mandatory and optional inputs
        self._validate_inputs(inputs)

        return self.blender.process(
            inputs = inputs,
            output_size = self.cpu_render_context.output_size,
            mix_weight = mix_weight,
            dtype = dtype,
            **kwargs
        )
    
    # # TODO: Should this be in the '_BlenderCPU' instead (?)
    # # TODO: This is not working yet, refactor
    # def process_multiple_inputs(
    #     self,
    #     inputs: dict[str, np.ndarray],
    #     # inputs: list[np.ndarray],
    #     output_size: Union[tuple[int, int], None],
    #     mix_weights: Union[list[float], float] = 1.0,
    #     dtype: Union[np.dtype, None] = None,
    #     **kwargs
    # ) -> np.ndarray:
    #     """
    #     Blend all the `inputs` provided, one after another,
    #     applying the `mix_weight` provided, and forcing the
    #     result to the `dtype` if provided.

    #     The `mix_weight` can be a single float value, that 
    #     will be used for all the mixings, or a list of as
    #     many float values as `inputs` received, to be 
    #     applied individually to each mixing.
    #     """
    #     return self.blender.process_multiple_inputs(
    #         inputs = inputs,
    #         output_size = output_size,
    #         mix_weights = mix_weights,
    #         dtype = dtype,
    #         **kwargs
    #     )

class _BlenderCoreCPU(_NodeCPU, metaclass = SingletonABCMeta):
    """
    *Singleton class*

    *For internal use only*
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []

    def process(
        self,
        inputs: dict[str, np.ndarray],
        output_size: Union[tuple[int, int], None],
        mix_weight: float = 1.0,
        dtype: Union[np.dtype, None] = None,
        **kwargs
    # TODO: Is this the expected type (?)
    # TODO: What about OpenGL textures (?)
    ) -> np.ndarray:
        """
        Blend the `base_input` provided with the also given
        `overlay_input`, using the `mix_weight` to obtain
        a single output as the result.

        The inputs should be received in uint8 format in
        the [0, 255] range, then processed with the specific
        processing method, and then turned back again to the
        uint8 format and [0, 255] range.
        """
        ParameterValidator.validate_mandatory_number_between('mix_weight', mix_weight, 0.0, 1.0)

        self._validate_inputs(inputs)

        # TODO: We should always receive the inputs as
        # uint8 in [0, 255] range

        # We will force the same size for the inputs
        """
        By now the strategy we are applying is to scale
        all the inputs to the size of the biggest one.
        TODO: Create more strategies and add a setting
        to be able to choose it
        """
        base_input = inputs['base_input']
        overlay_input = inputs['overlay_input']

        base_input, overlay_input = InputHandler.scale_to_biggest(
            inputs = [base_input, overlay_input]
        )

        # TODO: We have to make sure the sizes are the
        # same, or force them or do something...
        _validate_inputs_and_mix_weights(
            inputs = [base_input, overlay_input],
            mix_weights = [mix_weight]
        )

        # Transform to numpy arrays if textures received
        # base_input = (
        #     texture_to_frame(base_input, do_include_alpha = True)
        #     if PythonValidator.is_instance_of(base_input, 'moderngl.Texture') else
        #     base_input
        # )

        # overlay_input = (
        #     texture_to_frame(overlay_input, do_include_alpha = True)
        #     if PythonValidator.is_instance_of(overlay_input, 'moderngl.Texture') else
        #     overlay_input
        # )

        dtype = (
            base_input.dtype
            if dtype is None else
            dtype
        )

        # TODO: Size of 'base_input' and 'overlay_input' must
        # be the same, and the 'base_input' should be the size
        if mix_weight == 0:
            # We want to affect 0% to the base, so we return the
            # base directly instead of calculating, but... Y_Y
            return base_input

        # 1. Blend with the specific internal process
        blended = self._blend(
            base_input = base_input,
            overlay_input = overlay_input,
            output_size = output_size,
            **kwargs
        )

        # 2. Apply mix_weight to determine the percentage to affect
        blended = (
            base_input * (1 - mix_weight) + blended * mix_weight
            if mix_weight < 1.0 else
            blended
        )

        # Always turn back to uint8 in [0, 255] range
        blended = _ensure_uint8(blended)

        return blended.astype(dtype)

    # def process_multiple_inputs(
    #     self,
    #     inputs: list[np.ndarray],
    #     output_size: Union[tuple[int, int], None],
    #     mix_weights: Union[list[float], float] = 1.0,
    #     dtype: Union[np.dtype, None] = None,
    #     **kwargs
    # ) -> np.ndarray:
    #     """
    #     Blend all the `inputs` provided, one after another,
    #     applying the `mix_weight` provided, and forcing the
    #     result to the `dtype` if provided.

    #     The `mix_weight` can be a single float value, that 
    #     will be used for all the mixings, or a list of as
    #     many float values as `inputs` received, to be 
    #     applied individually to each mixing.
    #     """
    #     inputs = InputHandler.scale_to_biggest(
    #         inputs = inputs
    #     )

    #     _validate_inputs_and_mix_weights(
    #         inputs = inputs,
    #         mix_weights = mix_weights
    #     )
        
    #     # We process all the 'inputs' as 'base' and 'overlay'
    #     # and accumulate the result
    #     dtype = (
    #         inputs[0].dtype
    #         if dtype is None else
    #         dtype
    #     )

    #     # Use the first one as the base
    #     base = inputs[0]

    #     # TODO: How do we handle the additional parameters that
    #     # could be an array? Maybe if it is an array, check that
    #     # the number of elements is the same as the number of
    #     # inputs, and if a single value just use it...

    #     for i in range(1, len(inputs)):
    #         overlay = inputs[i]
    #         mix_weight = mix_weights[i]

    #         # TODO: If we make a lot of different operations and
    #         # force the [0, 255] range in each of them, we could
    #         # have an accumulated error... Think if we need to
    #         # refactor this
    #         base = self.process(
    #             inputs = {
    #                 'base_input': base,
    #                 'overlay_input': overlay
    #             },
    #             output_size = output_size,
    #             mix_weight = mix_weight
    #         )

    #     # Result is always forced to uint8 and [0, 255] when
    #     # processed by pairs, so we don't need  to do it
    #     # TODO: Maybe we can avoid 'dtype' parameter as it
    #     # is not useful at all...
    #     return base.astype(dtype)
    
    @abstractmethod
    def _blend(
        self,
        base_input: np.ndarray,
        overlay_input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        **kwargs
    # TODO: Is this the expected type (?)
    # TODO: What about OpenGL textures (?)
    ) -> np.ndarray:
        """
        *For internal use only*

        *This method must be overwritten by the specific
        classes*

        The internal process to blend and mix the provided
        `base_input` and `overlay_input`.

        This method should not force uint8 nor [0, 255]
        range by itself as it would be done in the 'blend'
        main method.
        """
        pass


"""
To build scenes.
"""
from yta_editor_nodes_cpu.compositor.displacement_with_rotation import DisplacementWithRotationNodeCompositorCPU


__all__ = [
    'DisplacementWithRotationNodeCompositorCPU'
]


from yta_editor_nodes_cpu.abstract import _NodeCPU
from yta_programming.singleton import SingletonABCMeta
from abc import abstractmethod
from typing import Union

import numpy as np


class _NodeProcessorCPU(
    _NodeCPU
):
    """
    *Abstract class*

    *For internal use only*

    Node class, which is to implement specific
    parameters when instantiated and also a specific
    node processor that will process the `input`.
    """

    mandatory_inputs = []
    optional_inputs = []

    def __init__(
        self,
        node_processor: '_NodeProcessorCoreCPU',
        cpu_render_context: 'CPURenderContext'
    ):
        # TODO: Validate that is '_NodeProcessorCoreCPU' or child
        self.node_processor: '_NodeProcessorCoreCPU' = node_processor
        """
        The singleton instance of the processor that will
        process the input and return the output.
        """
        self.cpu_render_context: 'CPURenderContext' = cpu_render_context
        """
        The render context for this CPU node processor.
        """

    # TODO: Can I refactor somewhow to make it have a
    # common validation (?)
    @abstractmethod
    def process(
        self,
        inputs: dict[str, np.ndarray],
        evaluation_context: 'EvaluationContext',
        # TODO: Should we set the 'factor' here to be able
        # to modify it if it is different than None (?)
        **kwargs
    ) -> np.ndarray:
        """
        Process the `inputs` provided by using the node
        processor instance associated to this node and
        return the output.

        This method must be defined for each class.
        """
        pass
        # ParameterValidator.validate_dict_of('inputs', inputs, np.ndarray)

        # # Validate mandatory and optional inputs
        # self._validate_inputs(inputs)

        # return self.process(
        #     inputs = inputs,
        #     output_size = output_size,
        #     **kwargs
        # )

class _NodeProcessorCoreCPU(metaclass = SingletonABCMeta):
    """
    *Singleton class*

    *For internal use only*

    Class to represent a node processor that uses CPU
    to transform the input. This class will be called
    internally by the specific nodes to process the
    inputs.
    """

    # TODO: Just code and the same attributes that the
    # GPU version also has
    @abstractmethod
    def process(
        self,
        input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided 'input' and transform it by
        using the code that is defined here.
        """
        # TODO: Specific attributes can be received as
        # **kwargs to modify the specific process
        pass


from yta_editor_nodes_cpu.abstract import _NodeCPU
from yta_programming.singleton import SingletonABCMeta
from abc import ABC, abstractmethod
from typing import Union

import numpy as np


class _NodeCompositorCPU(_NodeCPU):
    """
    *Abstract class*

    *For internal use only*

    Class to represent a node compositor that uses CPU
    to position the input in a specific place.

    Node class, which is to implement specific
    parameters when instantiated and also a specific
    node compositor that will process the `input`.

    This _NodeCompositor is to be able to make a composition
    with different inputs. For example, placing a specific
    input in a position, with a rotation, over another input
    that acts as the base.
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []

    def __init__(
        self,
        node_compositor: '_NodeCompositorCoreCPU',
        cpu_render_context: 'CPURenderContext'
    ):
        # TODO: Validate that is '_NodeCompositorCoreCPU' or child
        self.node_compositor: '_NodeCompositorCoreCPU' = node_compositor
        """
        The singleton instance of the compositor that will
        process the input and return the output.
        """
        self.cpu_render_context: 'CPURenderContext' = cpu_render_context
        """
        The render context for this CPU node processor.
        """

    def process(
        self,
        inputs: dict[str, np.ndarray],
        evaluation_context: 'EvaluationContext',
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided 'input' and transform it by
        using the code that is defined here.
        """
        return self.node_compositor.process(
            base_input = inputs['base_input'],
            overlay_input = inputs['overlay_input'],
            output_size = self.cpu_render_context.output_size
        )

class _NodeCompositorCoreCPU(metaclass = SingletonABCMeta):
    """
    *Singleton class*

    *For internal use only*

    This class is to be able to make a composition with
    different inputs. For example, placing an input in a
    specific position, with a rotation, over another
    input that acts as the base.

    This class will be called internally by the specific
    nodes to make the composition.
    """

    # def __init__(
    #     self,
    #     **kwargs
    # ):
    #     pass

    # def __reinit__(
    #     self,
    #     **kwargs
    # ):
    #     pass
    
    # TODO: Just code and the same attributes that the
    # GPU version also has
    @abstractmethod
    def process(
        self,
        base_input: np.ndarray,
        overlay_input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided 'input' and transform it by
        using the code that is defined here.
        """
        # TODO: Specific attributes can be received as
        # **kwargs to modify the specific process
        pass


"""
TODO: This module doesn't use 't' but 'progress'
so it is not a child of 'processor.video', maybe
we should move it to be 'processor.transitions'
instead of 'processor.video.transitions'... (?)
"""
from yta_editor_nodes_cpu.processor.video.transitions.slide import SlideTransitionProcessorCPU


__all__ = [
    'SlideTransitionProcessorCPU'
]


from yta_editor_nodes_cpu.processor.abstract import _NodeProcessorCPU, _NodeProcessorCoreCPU
from yta_editor_utils.texture import TextureUtils
from typing import Union

import numpy as np


class SelectionMaskNodeProcessorCPU(_NodeProcessorCPU):
    """
    The node to select the content by using a mask
    to determine if the pixel must be applied or
    not.

    This class can be instantiated many different
    times with different parameters, but will always
    call the same `Singleton` node processor instance
    to process the `input`.
    """

    mandatory_inputs = ['original_inpupt', 'processed_input', 'selection_mask_input']
    optional_inputs = []
    
    def __init__(
        self,
        cpu_render_context: 'CPURenderContext'
    ):
        super().__init__(
            node_processor = _SelectionMaskProcessorCoreCPU(),
            cpu_render_context = cpu_render_context
        )

    def process(
        self,
        inputs: dict[str, np.ndarray],
        evaluation_context: 'EvaluationContext',
        **kwargs
    ) -> np.ndarray:
        """
        Process the `input` provided by using the node
        processor instance associated to this node and
        return the output.
        """
        return self.node_processor.process(
            original_input = inputs['original_input'],
            processed_input = inputs['processed_input'],
            selection_mask_input = inputs['selection_mask_input'],
            output_size = self.cpu_render_context.output_size
        )

class _SelectionMaskProcessorCoreCPU(_NodeProcessorCoreCPU):
    """
    *For internal use only*
    
    Class to use a mask selection (from which we will
    determine if the pixel must be applied or not) to
    apply the `processed_input` on the `original_input`.
    """

    def process(
        self,
        original_input: np.ndarray,
        processed_input: np.ndarray,
        selection_mask_input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        **kwargs
    ):
        """
        Apply the `selection_mask` provided to the also
        given `original` and `processed` nuumpy arrays to
        obtain the processed one but affected only as the
        selection mask says.

        The input are processed as float32, with float 
        precission, to be able to calculate properly,
        and then returned to uint8 [0, 255] values (the
        ones our OpenGL is able to handle with the 'f1'
        dtype and the sampler2d uniforms).
        """
        # We force to have float precission for the calculations
        original_input = TextureUtils.numpy_to_float32(original_input)
        processed_input = TextureUtils.numpy_to_float32(processed_input)
        selection_mask_input = TextureUtils.numpy_to_float32(selection_mask_input)

        # We need a 3D or 4D mask
        selection_mask_input = (
            np.expand_dims(selection_mask_input, axis = -1)
            if selection_mask_input.ndim == 2 else
            selection_mask_input
        )

        selection_mask_input = (
            np.repeat(
                a = selection_mask_input,
                repeats = original_input.shape[-1],
                axis = -1
            )
            if (
                selection_mask_input.shape[-1] == 1 and
                original_input.shape[-1] in (3, 4)
            ) else
            selection_mask_input
        )

        # Mix with the selection mask
        final = original_input * (1.0 - selection_mask_input) + processed_input * selection_mask_input

        return TextureUtils.numpy_to_uint8(final)


"""
This is the nodes module, in which we have
all the classes that make the concept work.

This module is specific to process with CPU.

You have access to a node that can be
instantiated as many times as you need, with
different parameters, that will call a 
singleton instance in the backend to process
the inputs.

Note for the developer:
- Make sure you use 'f1' dtype with [0, 255]
values (np.uint8) for the inputs, using the
float precission sometimes but always returning
those inputs for the textures because those are 
the ones we can handle with our OpenGL system.

IMPORTANT NOTES:
- The `input` for the CPU nodes must be `float32`
and the output will be given as a `float32`.
"""


from yta_editor_nodes_cpu.processor.video.abstract import _VideoNodeProcessorCoreCPU, _VideoNodeProcessorCPU
from yta_programming.decorators.requires_dependency import requires_dependency
from typing import Union

import numpy as np


class WavingFrameVideoNodeProcessorCPU(_VideoNodeProcessorCPU):
    """
    The node to modify the brightness of an input
    by using the CPU.

    This class can be instantiated many different
    times with different parameters, but will always
    call the same `Singleton` node processor instance
    to process the `input`.
    """

    mandatory_inputs = ['base_input']
    optional_inputs = []
    
    def __init__(
        self,
        cpu_render_context: 'CPURenderContext'
    ):
        super().__init__(
            node_processor = WavingFrameVideoNodeProcessorCoreCPU(),
            cpu_render_context = cpu_render_context
        )

    def process(
        self,
        inputs: dict[str, np.ndarray],
        evaluation_context: 'EvaluationContext',
        amplitude: float = 0.05,
        frequency: float = 10.0,
        speed: float = 2.0,
        do_use_transparent_pixels: bool = False,
        **kwargs
    ) -> np.ndarray:
        """
        Process the `input` provided by using the node
        processor instance associated to this node and
        return the output.
        """
        return self.node_processor.process(
            input = inputs['base_input'],
            output_size = self.cpu_render_context.output_size,
            t = float(evaluation_context.t),
            amplitude = amplitude,
            frequency = frequency,
            speed = speed,
            do_use_transparent_pixels = do_use_transparent_pixels,
        )

class WavingFrameVideoNodeProcessorCoreCPU(_VideoNodeProcessorCoreCPU):
    """
    *Optional `opencv-python` (imported as `cv2`) library is required*
    
    Just an example of a specific class that is a node
    processor that uses CPU.
    """

    @requires_dependency('cv2', 'yta_editor_nodes_cpu', 'opencv-python')
    def process(
        self,
        input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        t: float,
        amplitude: float = 0.05,
        frequency: float = 10.0,
        speed: float = 2.0,
        do_use_transparent_pixels: bool = False,
        **kwargs
    ) -> np.ndarray:
        """
        Process the provided 'input' and transform it by
        using the code that is defined here.
        """
        import numpy as np
        import cv2  # opcional, solo para redimensionar/interpolar

        h, w, _ = input.shape

        # if input.shape[2] == 3:
        #     # Add alpha if needed
        #     input = cv2.cvtColor(input, cv2.COLOR_BRG2BRGA)

        # UV coordinates [0, 1]
        x = np.linspace(0, 1, w, dtype = np.float32)
        y = np.linspace(0, 1, h, dtype = np.float32)
        xv, yv = np.meshgrid(x, y)
        # Flip vertically to match OpenGL (GPU) orientation
        yv = 1.0 - yv  

        wave = np.sin(xv * frequency + t * speed) * amplitude

        # UV coordinates displacement
        yv_new = yv + wave

        # UV coordinates to pixels
        map_x = (xv * (w - 1)).astype(np.float32)
        # Invert back for image space (to be similar to GPU effect)
        map_y = ((1.0 - yv_new) * (h - 1)).astype(np.float32)

        output = cv2.remap(
            input,
            map_x,
            map_y,
            interpolation = cv2.INTER_LINEAR,
            borderMode = cv2.BORDER_REFLECT
        )

        if do_use_transparent_pixels:
            mask = (yv_new < 0.0) | (yv_new > 1.0)
            output[mask, 3] = 0

        return output


from yta_editor_nodes_cpu.blender.abstract import _BlenderCoreCPU, _BlenderCPU
from yta_editor_utils.texture import TextureUtils
from typing import Union

import numpy as np


class AlphaBlenderCPU(_BlenderCPU):
    """
    Blend the different inputs by applying the
    normal and most famous alpha blending and the
    most typical used in video editing.

    This blender will use the alpha channel of the 
    overlay input, multiplied by the `blend_strength`
    parameter provided, to use it as the mixer factor
    between the base and the overlay inputs.
    """

    def __init__(
        self,
        cpu_render_context: 'CPURenderContext'
    ):
        super().__init__(
            blender = _AlphaBlenderCoreCPU(),
            cpu_render_context = cpu_render_context
        )

class _AlphaBlenderCoreCPU(_BlenderCoreCPU):
    """
    *Singleton class*

    *For internal use only*

    Blend the different inputs by applying the
    normal and most famous alpha blending and the
    most typical used in video editing.

    This blender will use the alpha channel of the 
    overlay input, multiplied by the `blend_strength`
    parameter provided, to use it as the mixer factor
    between the base and the overlay inputs.
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []
    
    def process(
        self,
        inputs: dict[str, np.ndarray],
        output_size: Union[tuple[int, int], None],
        mix_weight: float = 1.0,
        dtype: Union[np.dtype, None] = None,
        blend_strength: float = 0.5,
        **kwargs
    ) -> np.ndarray:
        # TODO: Maybe obtain the 'output_size' from
        # the 'base_input' if not defined...
        return super().process(
            inputs = inputs,
            output_size = output_size,
            mix_weight = mix_weight,
            dtype = dtype,
            blend_strength = blend_strength
        )
    
    def _blend(
        self,
        base_input: np.ndarray,
        overlay_input: np.ndarray,
        output_size: Union[tuple[int, int], None],
        blend_strength: float = 1.0,
        **kwargs
    ) -> np.ndarray:
        """
        *For internal use only*

        The internal process to blend and mix the provided
        `base_input` and `overlay_input`.

        (!) This method should not force uint8 nor [0, 255]
        range by itself as it would be done in the 'blend'
        main method.
        """
        # We force to have float precission for the calculations
        base_input = TextureUtils.numpy_to_float32(base_input)
        overlay_input = TextureUtils.numpy_to_float32(overlay_input)

        base_rgb, base_a = base_input[..., :3], base_input[..., 3:4]
        overlay_rgb, overlay_a = overlay_input[..., :3], overlay_input[..., 3:4]

        # Compound alpha
        alpha = overlay_a * blend_strength

        # Equivalent to mix(base, overlay, alpha) in glsl
        out_rgb = base_rgb * (1.0 - alpha) + overlay_rgb * alpha
        out_a = base_a * (1.0 - alpha) + overlay_a * alpha

        return TextureUtils.numpy_to_uint8(np.concatenate(
            [out_rgb, out_a],
            axis = -1
        ))